package store

import (
	"strings"

	"github.com/dgraph-io/badger/v3"
	"github.com/vmihailenco/msgpack/v5"
)

type badgerStore struct {
	db *badger.DB
}

// NewBadger creates a new store from a badger database
func NewBadger(db *badger.DB) Store {
	return badgerStore{db}
}

func (b badgerStore) Set(key string, val interface{}) error {
	return b.db.Update(func(txn *badger.Txn) error {
		data, err := msgpack.Marshal(val)
		if err != nil {
			return err
		}
		return txn.Set([]byte(key), data)
	})
}

func (b badgerStore) Get(key string, out interface{}) error {
	return b.db.View(func(txn *badger.Txn) error {
		item, err := txn.Get([]byte(key))
		if err != nil {
			return err
		}

		return item.Value(func(val []byte) error {
			return msgpack.Unmarshal(val, out)
		})
	})
}

func (b badgerStore) GetItem(key string) (*Item, error) {
	out := &Item{}
	err := b.db.View(func(txn *badger.Txn) error {
		item, err := txn.Get([]byte(key))
		if err != nil {
			return err
		}

		key := string(item.KeyCopy(nil))
		data, err := item.ValueCopy(nil)
		if err != nil {
			return err
		}

		out.Key = key
		out.Data = data

		return nil
	})
	return out, err
}

func (b badgerStore) Delete(key string) error {
	return b.db.Update(func(txn *badger.Txn) error {
		return txn.Delete([]byte(key))
	})
}

func (b badgerStore) Modify(key string, modify ModifyFunc) error {
	item, err := b.GetItem(key)
	if err != nil {
		return err
	}

	newVal, err := modify(item)
	if err != nil {
		return err
	}

	return b.Set(key, newVal)
}

func (b badgerStore) Iter() Iter {
	out := make(chan *Item, 10)
	go func() {
		b.db.View(func(txn *badger.Txn) error {
			opts := badger.DefaultIteratorOptions
			it := txn.NewIterator(opts)
			defer it.Close()
			for it.Rewind(); it.Valid(); it.Next() {
				item := it.Item()
				key := string(item.KeyCopy(nil))
				valData, err := item.ValueCopy(nil)
				if err != nil {
					continue
				}

				out <- &Item{Key: key, Data: valData}
			}
			close(out)
			return nil
		})
	}()
	return out
}

func (b badgerStore) Bucket(name string) Bucket {
	return badgerBucket{b, name}
}

type badgerBucket struct {
	store badgerStore
	name  string
}

func (b badgerBucket) Name() string {
	return b.name
}

func (b badgerBucket) Set(key string, val interface{}) error {
	return b.store.Set(b.name+"\u001f"+key, val)
}

func (b badgerBucket) Get(key string, out interface{}) error {
	return b.store.Get(b.name+"\u001f"+key, out)
}

func (b badgerBucket) GetItem(key string) (*Item, error) {
	return b.store.GetItem(b.name + "\u001f" + key)
}

func (b badgerBucket) Delete(key string) error {
	return b.store.Delete(b.name + "\u001f" + key)
}

func (b badgerBucket) Modify(key string, modify ModifyFunc) error {
	return b.store.Modify(b.name+"\u001f"+key, modify)
}

func (b badgerBucket) Iter() Iter {
	out := make(chan *Item, 10)
	go func() {
		b.store.db.View(func(txn *badger.Txn) error {
			opts := badger.DefaultIteratorOptions
			opts.Prefix = []byte(b.name + "\u001f")
			it := txn.NewIterator(opts)
			defer it.Close()
			for it.Rewind(); it.Valid(); it.Next() {
				item := it.Item()
				key := string(item.KeyCopy(nil))
				key = strings.TrimPrefix(key, b.name+"\u001f")
				valData, err := item.ValueCopy(nil)
				if err != nil {
					continue
				}

				out <- &Item{Key: key, Data: valData}
			}
			close(out)
			return nil
		})
	}()
	return out
}
